<?php
// Copyright (c) 2005-2010, RedIRIS. All Rights Reserved.
//
// You may distribute under the terms of the GNU General Public License,
// as specified in the LICENSE file that was shipped with this distribution

require_once 'crypt.php';
require_once 'DBInterface.php';
require_once 'DBManager.php';

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS abstractPoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

abstract class abstractPoA {
	var $local_site;
	var $cfg; // Configuration parameters
	var $globalExpireTime;
	var $enforcing = true;
	var $send_request = false;
	var $skip_redirection = false;
	var $Lcook_name = "PAPILcook_";
	var $assertion;
	var $authz;
	var $dbManager;
        var $id_string = "http";
        var $send_id = true;

	abstract function set_parameters();
	abstract function log($message);
	abstract function send_Request();
	abstract function check_Access();

	function abstractPoA($local_site) {
		$this->local_site = $local_site;

		if (!$this->set_parameters()) {
			trigger_error("Errors during $this configuration", E_USER_ERROR);
			return null;
		}

		if (!$this->check_parameters()) {
			trigger_error("Required parameters missing for $local_site", E_USER_ERROR);
			return null;
		}

		// Get the Simetric Key
		$fp1 = fopen($this->cfg["LKEY_File"], "r");
		$this->lkey = fread($fp1, 8192);
		fclose($fp1);

		// Get the Public Key
		$this->get_public_key();

		// Set the Lcook name and avoid problems with dots in array keys
		$this->Lcook_name = $this->Lcook_name.$this->local_site;
		if (strpos ($this->Lcook_name, '.') != false) {
			$this->Lcook_name = str_replace ( '.', '_' , $this->Lcook_name );
		}

                if (isset($_SERVER['HTTPS'])) $this->id_string .= "s";
                $this->id_string .= "://".$_SERVER['SERVER_NAME'].$this->cfg["Location"];

		// Get DB driver
		$this->dbManager = DBManager::getDBDriver($this->cfg['DB_Type']);
	}

	function getHomeLocatorIdentifier() {
		return $this->cfg['Home_Locator'];
	}

	function setHomeLocatorIdentifier($hli) {
		$this->cfg['Home_Locator'] = $hli;
	}

	function isEnforcing() {
		return $this->enforcing !== false;
	}

	function setEnforcing($enforce) {
		$this->enforcing = $enforce;
	}

        function sendID($send) {
            $this->send_id = $send;
        }

        function setID($id) {
            $this->id_string = $id;
        }

	// Taken from php.net to replace php 5 functions
	function __http_build_query($formdata, $numeric_prefix) {
		$arr = array ();
		foreach ($formdata as $key => $val)
			$arr[] = urlencode($numeric_prefix . $key) . "=" . urlencode($val);
		return implode($arr, "&");
	}

	// Taken from PHP OpenID library 2.1.2
	function params_from_string($str) {
		$chunks = explode('&', $str);
		$data = array();

		foreach ($chunks as $chunk) {
			$parts = explode('=', $chunk, 2);

			if (count($parts) != 2)
				continue;

			list($k, $v) = $parts;
			$data[$k] = urldecode($v);
		}

		return $data;
	}
	
	function access_Status() {
		$this->skip_redirection = true;
		list($code, $this->assertion) = $this->papi_Engine();
		$this->skip_redirection = false;
		$userData = $this->create_Hash($code, $this->assertion);
		return $userData;
	}

	function papi_Engine() {
		if(isset($this->cfg["Pass_Pattern"])) {
			$pattern = $this->pass_pattern();
			if($pattern){
				$this->log("The request matched the pass pattern '$pattern'");
				$userAssertion = $pattern;
				return array(2, $userAssertion,null);
			}
		}

		// check if src IP is allowed or denied
		if (!$this->isEnforcing()) {
			list($result,$filter) = $this->test_IP_Filters();
			switch ($result) {
				case 1:
					$this->log("The client matched the allowed IP pattern");
					return array(2, $filter, null);
				case -1:
					$this->log("The client matched the denied IP pattern");
					return array(0, "", $this->cfg['Not_Auth_Error_File']);
			}
		}

		$action = array_key_exists('ACTION', $_REQUEST) ? $_REQUEST['ACTION'] : "";
		$auth = (isset($_COOKIE[$this->Lcook_name])) ? $this->test_Lcook($_COOKIE[$this->Lcook_name]) : 0;
		$auth = (is_array($auth)) ? $auth[0] : $auth;
		if ($action == "CHECKED" and $auth != 1) { // GpoA/AS Response
			$data = $_REQUEST['DATA'];
			list ($valid, $key, $userAssertion) = $this->test_Response($data, $this->pub_key);
			$req_s = $this->load_Request($key);
			if (!$req_s) {
				if ($this->skip_redirection) { return array(2, "", null); }
			}
			$this->delete_Request($key);
			if ($valid) { // GpoA/AS has validated the request. Check the filters
				if (!$this->test_Filters($userAssertion)) { // Data does not match  the filters
					$error_message = "Assertion $userAssertion does not match filters";
					$this->log($error_message);
					# If AutoPoA automatically redirect error page
					# If PoA return userAssertion
					return array(0, $userAssertion, $this->cfg['Not_Auth_Error_File']);
				}

				// not filtered, continue!
				if ($this->skip_redirection) { return array(1, $userAssertion,null); }

				$Lcook_cookie = $this->gen_Lcook($userAssertion);
				if (setcookie($this->Lcook_name, $Lcook_cookie, 0, $this->cfg["Location"], $this->cfg["Cookie_Domain"], 0)) {
					$_COOKIE[$this->Lcook_name] = $Lcook_cookie;
					$req = unserialize($req_s);

					if (!$this->send_request) {
						$_GET = $req[0];
						$_POST = $req[1];
						$_REQUEST = $req[2];
						$_SERVER["QUERY_STRING"] = $req[3];
						$_SERVER["REQUEST_METHOD"] = $req[4];
						return array(1, $userAssertion,null);
					} else {
						if (strcmp($req[4], "GET") == 0) {
							if (isset($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == "on")
								$sch = "https://";
							else
								$sch = "http://";

							$ru = $sch.$_SERVER["SERVER_NAME"].":".$_SERVER["SERVER_PORT"].
								$_SERVER["SCRIPT_NAME"]."?".$req[3];
							header("HTTP/1.1 302 Found");
							header("Location: $ru");
						} else {
							$form_inputs = "";

							foreach ($this->params_from_string($req[5]) as $k => $v)
								$form_inputs .= "<input name='".$k."' type='hidden' value='".$v."' />";

							$form_inputs .= "<input type='submit' value='Please, continue...' />";
							print("<html>".
								"<head><title>phpPoA transaction in progress...</title></head>".
								"<body onload='document.forms[0].submit();'>".
								"<form action='".$_SERVER["SCRIPT_NAME"]."' method='post'>".$form_inputs."</form>".
								"<script>".
								"var elements = document.forms[0].elements;".
								"for (var i = 0; i < elements.length; i++) {".
								"  elements[i].style.display = \"none\";".
								"}".
								"</script>".
								"</body>".
								"</html>");
						}

						exit;
					}
				} else {
					// Cookie has not been sent OK
					$error_message = "Cookie has not been correctly sent";
					$this->log($error_message);
					# If AutoPoA automatically redirect error page
					# If PoA return userAssertion
					return array(-1, $userAssertion, $this->cfg['Cookie_Error_File']);
				}
			} else {
				// GpoA/AS has not validated the request. Do not allow access and send ERROR page.
				$error_message = "GpoA/AS has not validated the request";
				$this->log($error_message);
				# If AutoPoA automatically redirect error page
				# If PoA return userAssertion
				return array(0, $userAssertion, $this->cfg['Not_Auth_Error_File']);
			}
		} else { // Direct Access from the client browser
			$Lcook = array_key_exists($this->Lcook_name, $_COOKIE) ? $_COOKIE[$this->Lcook_name] : "";
			if ($Lcook == "") {
				$error_message = "Cookie ".$this->Lcook_name." not found in request";
				$this->log($error_message);
				if ($this->skip_redirection) { return array(0, "", $this->cfg['Not_Auth_Error_File']); }
				setcookie($this->Lcook_name,"",time()-3600);
				return $this->send_Request ();
			} else {
				// Check if Lcook is correct
				list ($res, $userAssertion) = $this->test_Lcook($Lcook);
				switch ($res) {
					case 0 :
						// User not authorized
						$error_message = "User not authorized.";
						$this->log($error_message);
						# If AutoPoA automatically redirect error page
						# If PoA return userAssertion
						return array(0, $userAssertion, $this->cfg['Not_Auth_Error_File']);
					case 1 :
						// Cookie Lcook is correct: Update cookie
						$Lcook_cookie = $this->gen_Lcook($userAssertion);
						if (setcookie($this->Lcook_name, $Lcook_cookie, 0, $this->cfg["Location"], $this->cfg["Cookie_Domain"], 0)) {
							$_COOKIE[$this->Lcook_name] = $Lcook_cookie;
							// If no request found, assume this is a visit to a previously stored URL
                        				// (reloaded by the user or in some history list at the browser)
                                			// Rebuild the request without the ACTION and DATA parameters
                                			if (isset ($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on') {
								$sch = 'https://'; 
							} else {
								$sch = 'http://';
							}
							$re[] = "/ACTION=[^&]*&?/";
							$re[] = "/DATA=[^&]*&?/";
							$location = $sch.$_SERVER["SERVER_NAME"].":".$_SERVER["SERVER_PORT"].
									rtrim(preg_replace($re,"",$_SERVER["REQUEST_URI"]),"&?");
							if (isset($_GET["ACTION"]) || isset ($_GET["DATA"])) {
								header("Location: $location");
								exit;
							}
							return array(1, $userAssertion,null);
						} else {
							// Cookie has not been updated OK
							$error_message = "Cookie has not been correctly updated.";
							$this->log($error_message);
							# If AutoPoA automatically redirect error page
							# If PoA return userAssertion
							return array(1, $userAssertion, $this->cfg['Cookie_Error_File']);
						}
					case 2 :
						// Lcook has expired: Request to GpoA/AS
						if ($this->skip_redirection) { return array(0, "", $this->cfg['Not_Auth_Error_File']); }
						setcookie($this->Lcook_name,$_COOKIE[$this->Lcook_name],time()-3600);
						return $this->send_Request();
				} // Switch
			} // else ($Lcook != "")
		} // else ($action != "CHECKED")
	} // Function papi_Engine

	// Check current authentication status (JAAD-2008.12.26)
	/*function access_Status() {
		$auth = (isset($_COOKIE['Lcook'])) ? $this->test_Lcook($_COOKIE['Lcook']) : 0;
		$auth = (is_array($auth)) ? $auth[0] : $auth;
 		return ($auth == 1);
	}*/

	final function gen_Lcook($userAssertion) {
		$content = time().":".$this->globalExpireTime.":".$this->cfg["Location"].":".$this->local_site.":".$userAssertion;
		$Lcook = encrypt_AES($content, $this->lkey);
		return $Lcook;
	}

	final function test_Lcook($Lcook) {
		if ($Lcook == "")
			return array(0,null);
		// Return:
		// - 0: Lcook not valid.
		// - 1: Lcook valid.
		// - 2: Lcook expired.
		$newsource = decrypt_AES($Lcook, $this->lkey);
		$Lcook = explode(":", $newsource, 5);
		$timestamp = $Lcook[0];
		$this->globalExpireTime = $Lcook[1];
		$loc = $Lcook[2];
		$serv = $Lcook[3];
		$userAssertion = $Lcook[4]; // Code parameter of the Lcook is the userAssertion
		// Check the cookie
		if ($userAssertion == "") {
			$error_message = "test_Lcook() Function: Code parameter of Lcook is empty.";
			$this->log($error_message);
			return array(0,null);
		}
		if ($loc != $this->cfg["Location"]) {
			$error_message = "test_Lcook() Function: Location parameter of Lcook is not valid: $loc";
			$this->log($error_message);
			return array(0,$userAssertion);
		}
		if ($serv != $this->local_site) {
			$error_message = "test_Lcook() Function: Service_ID parameter of Lcook is not valid: $serv";
			$this->log($error_message);
			return array(0,$userAssertion);
		}
		$now = time();
		if (($this->globalExpireTime < $now) or ($timestamp + $this->cfg["Lcook_Timeout"] < $now)) {
			$error_message = "test_Lcook() Function: Lcook expired for Code $userAssertion.";
			$this->log($error_message);
			return array(2,$userAssertion);
		}
		if ($this->test_Filters($userAssertion)) {
			return array(1,$userAssertion);
		} else {
			$error_message = "test_Lcook() Function: Code $userAssertion matches rejection cookie filter.";
			$this->log($error_message);
			return array(0,$userAssertion);
		}
	}

	function test_Filters($userAssertion) {
		if (isset($this->cfg["PAPI_Filter_accept"])) {
			$accept = $this->cfg["PAPI_Filter_accept"];
		}
		if (isset($this->cfg["PAPI_Filter_reject"])) {
			$reject = $this->cfg["PAPI_Filter_reject"];
		}

		$local_match = $this->match_Filters($userAssertion, $accept, $reject);

		if ($local_match) {
			return 1;
		} else {
			$error_message = "test_Filters() Function: Code $userAssertion matches rejection cookie filter.";
			$this->log($error_message);
			return 0;
		}
	}

	// check if the source IP address matches any of the allowed or denied filters
	function test_IP_Filters() {
		// setup filtering criteria
                $search = array("/\./",
                                "/\.0/",
                                "/(:0){1,7}/",
                                "/^::/",
                                "/::$/");
                $replace = array("\.",
                                 ".\d{1,3}",
                                 "::",
                                 "(([0-9a-fA-F]{1,4})){1,7}\:",
                                 "(\:([0-9a-fA-F]{1,4})){1,7}");
                //$search = array(".",".0","::");
                //$replace = array("\.",".\d{1,3}","(\:([0-9a-fA-F]{1,4})){1,7}");
                $src_addr = $_SERVER['REMOTE_ADDR'];

		// check if there are IP filters
		if (isset($this->cfg["Allow_From"]) && !empty($this->cfg["Allow_From"])) {
                        $allowed_addr = $this->cfg["Allow_From"];
			if (is_array($allowed_addr)) {
	                        foreach ($allowed_addr as $ip) {
                	                $ip = preg_replace($search, $replace, $ip);
                        	        if (preg_match("/".$ip."/i", $src_addr)) {
                                	        $error_message = "test_IP_Filters(): Source IP address $src_addr matches allowed mask: $ip";
                                        	$this->log($error_message);
   	        	                        return array(1,$ip);
        	                        }
                        	}
                	} else {
				$ip = str_replace($search, $replace, $allowed_addr);
	                        if (preg_match("/".$ip."/", $src_addr)) {
        	                        $error_message = "test_IP_Filters(): Source IP address $src_addr matches allowed mask: $ip";
                	                $this->log($error_message);
                        	        return array(1,$ip);
	                        }
        	        }
                }
		if (isset($this->cfg["Deny_From"])) {
			$denied_addr = $this->cfg["Deny_From"];
			// denied IPs
			if (is_array($denied_addr)) {
				foreach ($denied_addr as $ip) {
					$ip = str_replace($search, $replace, $ip);
					if (preg_match("/".$ip."/", $src_addr)) {
						$error_message = "test_IP_Filters(): Source IP address $src_addr matches denied mask: $ip";
						$this->log($error_message);
						return array(-1,$ip);
					}
				}
			} else {
				$ip = str_replace($search, $replace, $denied_addr);
				if (preg_match("/".$ip."/", $src_addr)) {
					$error_message = "test_IP_Filters(): Source IP address $src_addr matches denied mask: $ip";
					$this->log($error_message);
					return array(-1,$ip);
				}
			}
		}
		return array(0,null);
	}

	function match_Filters($userAssertion, $accept, $reject) {
		// If there are no filters,
		// a implicit accept filter is defined: $accept == ".*"
		if (empty($accept) && empty($reject)) {
			return 1;
		} else if (empty($reject)) { // implicit rejection!
			// Accept
			if (!is_array($accept)) {
				$a[] = $accept;
				$accept = $a;
			}
			
			foreach ($accept as $filter) {
				if (preg_match('/'.$filter.'/', $userAssertion)) {
					$msg = "match_Filters(): assertion ".$userAssertion." matches acceptance filter: ".$filter.".";
					$this->log($msg);
					return 1;
				}
			}
		} else // $reject is not NULL
			{
			// Accept
			if (!is_array($accept)) {
				$a[] = $accept;
				$accept = $a;
			}

			foreach ($accept as $filter) {
				if ($filter) {
					if (preg_match('/'.$filter.'/', $userAssertion)) {
						$msg = "match_Filters(): assertion ".$userAssertion." matches acceptance filter: ".$filter.".";
						$this->log($msg);
						return 1;
					}
				}
			}

			// Reject
			if (!is_array($reject)) {
				$r[] = $reject;
				$reject = $r;
			}

			foreach ($reject as $filter) {
				if (preg_match('/'.$filter.'/', $userAssertion)) {
					$msg = "match_Filters(): assertion ".$userAssertion." matches rejection filter: ".$filter.".";
					$this->log($msg);
					return 0;
				}
			}
		}
		$msg = "match_Filters(): assertion ".$userAssertion." does not match any acceptance filter.";
		$this->log($msg);
		return 0;
	}

	function get_Acceptance_Filter() {
		return $this->cfg['PAPI_Filter_accept'];
	}

	function get_Rejection_Filter() {
		return $this->cfg['PAPI_Filter_reject'];
	}

	final function prepare_Request() {
		// marcoscm: Some clients don't rely *only* on $_REQUEST (2), but on $_GET (0), $_POST (1),
		// $_SERVER["QUERY_STRING"] (3), $_SERVER["REQUEST_METHOD"] (4) or and php://input (5)
		$buffer = array($_GET, $_POST, $_REQUEST, $_SERVER["QUERY_STRING"], $_SERVER["REQUEST_METHOD"],
			file_get_contents("php://input"));
		$req = serialize($buffer);
		list ($id, $key) = $this->save_Request($req);
		if (!$id) {
			$error_message = "It has not been possible to store the original request. Access not allowed";
			$this->log($error_message);
			return False;
		}
		if (isset ($_SERVER['HTTPS']) && $_SERVER['HTTPS'] == 'on') {
			$protocol = 'https://';
		} else {
			$protocol = 'http://';
		}

		//Check if the request hast to be sent to a GPoA or an AS
		$gpoaurl = $this->cfg["GPoA_URL"];

		// Direct Access to an AS
		if($gpoaurl == ""){
			//Format: URL = ASURL?ATTREQ=PoA/GPoAServiceID&PAPIPOAREF=value&PAPIPOAURL=value
			$ASURL = $protocol.$_SERVER['SERVER_NAME'].":".$_SERVER['SERVER_PORT'].$_SERVER['REQUEST_URI'];
			$parameters = array(
				'ATTREQ'=>'poaid',
				'PAPIPOAREF'=>$key,
				'PAPIPOAURL'=>$ASURL,
				);

			//Request to AS
			$request = $this->cfg["AS_URL"].'?'.$this->__http_build_query($parameters,"");
		} else {
			// Direct Access to a GPoA
			$PoAURL = $protocol.$_SERVER['SERVER_NAME'].":".$_SERVER['SERVER_PORT'].$_SERVER['REQUEST_URI'];
			$parameters = array (
				'ACTION' => 'CHECK',
				'DATA' => $key,
				'URL' => $PoAURL,
				);
                        if ($this->send_id) {
                                $parameters['PAPIOPOA'] = urlencode($this->id_string);
                        }
			if (isset($this->cfg['Home_Locator'])) {
				$parameters['PAPIHLI'] = $this->cfg['Home_Locator'];
			}

			// Request to GpoA
			$request = $this->cfg["GPoA_URL"].'?'.$this->__http_build_query($parameters,"");
		}
		return $request;
	}

	final function test_Response($data, $pub_key) {
		$newsource = "";
		// Decrypt the text
		$newsource = papi_openssl_decrypt($data, $pub_key, $this->cfg["Log"]);
		if (!$newsource) {
			// Cannot decrypt!
			$error_message = "test_Response() Function: Cannot decrypt response, check GPoA/AS public key.";
			$this->log($error_message);
			return array(0,null,null);
		}
		$response = explode(":", $newsource);
		$key = array_pop($response);
		$currentTime = array_pop($response);
		$this->globalExpireTime = array_pop($response);
		$userAssertion = implode(":", $response);
		if ($response[0] == "ERROR") { // The cookie is not valid. Not authorized User.
			$error_message = "test_Response() Function: The cookie is not valid. User not authorized.";
			$this->log($error_message);
			return array(0,$key,$userAssertion);
		} else { // Check the response (KEY,expiryTime,filters)
			// Check timestamps
			if ($this->globalExpireTime < time()) {
				// globally expired
				$error_message = "test_Response() function: response expired.";
				$this->log($error_message);
				return array(0,$key,$userAssertion);
			}
			if ($currentTime + $this->cfg["Lcook_Timeout"] < time()) {
				// expired
				$error_message = "test_Response() function: response expired.";
				$this->log($error_message);
				return array(0,$key,$userAssertion);
			}
			// Check the cookie
			if ($userAssertion == "") {
				// userAssertion is empty
				$error_message = "test_Response() Function: userAssertion is empty.";
				$this->log($error_message);
				return array(0,$key,$userAssertion);
			} else {
				// Check if the returned key from the GpoA/AS is stored in the DataBase
				$id = $this->dbManager->get_db($this->cfg);
				if (!$id) {
					$error_message = "test_Response() Cannot open DB.";
					$this->log($error_message);
					return array(0,$key,$userAssertion);
				} 
				elseif (!($this->dbManager->check_key($key, $id))) {
					// Key $key does not exist in Data Base $cfg[Req_DB]
					$error_message = "test_Response() Function: Key $key does not exist in Data Base ".$this->cfg['Request_DB'].".";
					$this->log($error_message);
                                        $this->dbManager->close_db($id);
					return array(0,$key,$userAssertion);
				}
                                $this->dbManager->close_db($id);
			}
			$this->log("test_Response() Function: Access granted for assertion: ".$userAssertion);
			return array(1,$key,$userAssertion);
		} // Check the response (KEY,expiryTime,filters)
	} // test_Response Function

	// DBM functions
	function save_Request($req) {
		// If DB does not exist, create it.
		$id = $this->dbManager->get_db($this->cfg);
		if (!$id) {
			$error_message = "save_Request() Function: Cannot open DB.";
			$this->log($error_message);
			return array($id,"");
		}
		$key = mt_rand();
		$this->dbManager->replace_key($key, $req, $id);
		$this->dbManager->close_db($id);
		return array($id,$key);
	}

	function load_Request($key) {
		$id = $this->dbManager->get_db($this->cfg);
		if (!$id) {
			$error_message = "load_Request() Cannot open DB.";
			$this->log($error_message);
			return "";
		}
		$req = false;
		if ($this->dbManager->check_key($key, $id)) {
			$req = $this->dbManager->fetch_key($key, $id);
		}
		$this->dbManager->close_db($id);
		return $req;
	}

	function delete_Request($key) {
		$id = $this->dbManager->get_db($this->cfg);
                $result = false;
		if (!$id) {
			$error_message = "delete_Request() Cannot open DB.";
			$this->log($error_message);
		}
		if ($this->dbManager->check_key($key, $id)) {
			$result = $this->dbManager->delete_key($key, $id);
		} else {
			$error_message = "delete_Request() Function: Key $key does not exist.";
			$this->log($error_message);
		}
		$this->dbManager->close_db($id);
		return $result;
	}

	final function create_Hash($code, $userAssertion) {
		$userData['PAPIAuthValue'] = $code;

		if ($code == 2) {
			$userData['PAPIPassPattern'] = $userAssertion;
		} else {
			// 128 is a security limit
			$assertComponents = explode("@", $userAssertion);
			$userData['PAPIASName'] = array_pop($assertComponents);
			$userData['PAPIAssertion'] = $userAssertion;
			$pureAssertion = implode("@", $assertComponents);
			// 128 is a security limit
			$elements = preg_split('/[\s]*,[\s]*/', $pureAssertion, 128);
			foreach ($elements as $element) {
				list ($key, $value) = preg_split('/[\s]*=[\s]*/', $element, 2);
				$userData[$key] = $value;
			}
		}

		return $userData;
	} // create_Hash()

	function check_parameters() {
		$ret = true;
		$gpoaurl = "";
		$asurl = "";

		$Params = Array (
			"Log",
			"Location", "Cookie_Domain",
			"LKEY_File", "Lcook_Timeout",
			"DB_Type"
		);
		foreach ($Params as $param) {
			if ( ! isset ($this->cfg[$param])) {
				$ret = false;
				$error_message = "set_Global_Parameters() Function: Parameter $param in $this->local_site does not exist.";
				$this->log($error_message);
			}
		}

		/* Check if the GPoA_URL or the AS_URL variables have been set up but one and only one of them
		 * can be configurated at the same time
		 * Firstly we check the local sectión
		 *
		 * 	- If only GpoA_URL or AS_URL are set up: OK
		 *	- If GPoA_URL and AS_URL are both set up: ERROR.
		 *	- If none of them is set up: check the global configuration in PAPI_Main section:
		 *			+ If GPoA_URL and AS_URL aren't set up: ERROR.
		 *			+ Only one is set up: OK
		 *			+ If GPo_AURL and AS_URL both are set up: ERROR
		*/
		if (isset($this->config_file[$this->local_site]["GPoA_URL"])) {
			$gpoaurl = $this->config_file[$this->local_site]["GPoA_URL"];
		}

		if (isset($this->config_file[$this->local_site]["AS_URL"])) {
			$asurl = $this->config_file[$this->local_site]["AS_URL"];
		}

		//Local section
		if ($gpoaurl and $asurl) {
			$ret = false;
			$error_message = "set_Global_Parameters() Function: Parameter GPoA_URL, or the AS_URL, in $this->local_site are set up at the same time. ";
			$this->log($error_message);
		} else if ($gpoaurl == "" and $asurl == "") {
			//PAPI_Main
			$gpoaurl = $this->config_file["PAPI_Main"]["GPoA_URL"];
			$asurl = $this->config_file["PAPI_Main"]["AS_URL"];
			if (($gpoaurl == "" and $asurl == "") or ($gpoaurl and $asurl)) {
				$ret = false;
				$error_message = "set_Global_Parameters() Function: Parameter GPoA_URL, or the AS_URL, in $this->local_site has to have a value, but only one them has to be set up. ";
				$this->log($error_message);
			}
		}
		return $ret;
	}

	function get_public_key() {
		$asurl = "";
		$gpoaurl = "";

		//LOCAL_SECTION: AS_URL or GPoA_URL are set up in the local section
		if (isset($this->config_file[$this->local_site]["AS_URL"])) {
			$asurl = $this->config_file[$this->local_site]["AS_URL"];
		} else {
			$gpoaurl = $this->config_file[$this->local_site]["GPoA_URL"];
		}


		if ($asurl) {	//Direct access to AS
			$fp2=fopen($this->cfg["AS_Pub_Key"],"r");
			$this->pub_key=fread($fp2,8192);
			fclose($fp2);
		} elseif ($gpoaurl) {	//Access to AS through GPoA
			$fp2 = fopen($this->cfg["GPoA_Pub_Key"],"r");
			$this->pub_key=fread($fp2,8192);
			fclose($fp2);
		} else {
			//MAIN_SECTION
			$asurl = $this->config_file["PAPI_Main"]["AS_URL"];
			$gpoaurl = $this->config_file["PAPI_Main"]["GPoA_URL"];

			if ($asurl) {	//Direct access to AS
				$fp2 = fopen($this->cfg["AS_Pub_Key"],"r");
				$this->pub_key=fread($fp2,8192);
				fclose($fp2);
			} elseif ($gpoaurl) { //Access to AS through GPoA
				$fp2 = fopen($this->cfg["GPoA_Pub_Key"],"r");
				$this->pub_key=fread($fp2,8192);
				fclose($fp2);
			}
		}
	}

	function pass_pattern() {
		$query = $this->getQuery();
		$patterns = explode(",", $this->cfg["Pass_Pattern"]);

		// Check if any pattern matches with the URI
		$uri = "$_SERVER[SERVER_NAME]$_SERVER[REQUEST_URI]";
		foreach ($patterns as $pattern) {
        	if (preg_match('/'.$pattern.'/',$uri)) {
	          return $pattern;
    	    }
      	}

		// Check if any pattern matches with the GET or POST params.
		foreach($query as $val) {
			foreach($patterns as $pattern) {
				if (preg_match('/' . $pattern  . '/', $val)){
					return $pattern;
				}
			}
		}
		return false;
	}

	function getQuery() {
		$data = explode("&", $_SERVER['QUERY_STRING']);

        if ($_SERVER['REQUEST_METHOD'] == 'POST') {
        	$str = file_get_contents('php://input');

	        if ($str === false) {
              	$post = array();
            } else {
              	$post = explode("&", $str);
            }
            $data = array_merge($data, $post);
        }
        return $data;
    }

	function getAuthorizationEngine() {
		return (isset($this->cfg['Authz_Engine'])) ? $this->cfg['Authz_Engine'] : "";
	}

    function isAuthorizedWrapper($user) {
    	if (!isset($this->assertion)) {
    		list($code, $this->assertion, $location) = $this->papi_Engine();
    	}
    	if (isset($this->cfg['Authz_Engine'])) {
    		include_once($this->cfg['Authz_Engine'].".php");
    		$authz = new $this->cfg['Authz_Engine']();
    		$authz->configure($this->cfg['Authz_Conf_File'], $this->local_site);
    		if ($authz->isAuthorized($user,$this->create_Hash(1,$this->assertion))) {
    			$this->log("isAuthorizedWrapper() Function: ".$this->assertion." authorized.");
    			return array(true, null);
    		}
    		$this->log("isAuthorizedWrapper() Function: ".$this->assertion." is not authorized.");
    		$this->log($authz->getError());
    		return array(false, $this->cfg['Not_Auth_Error_File']);
    	}
    	return array(true, null);
    }

	function authorizeWrapper($user, $force) {
    	if (!isset($this->assertion)) {
    		list($code, $this->assertion, $location) = $this->papi_Engine();
    	}
		if (isset($this->cfg['Authz_Engine'])) {
    		include_once($this->cfg['Authz_Engine'].".php");
    		$authz = new $this->cfg['Authz_Engine']();
    		$authz->configure($this->cfg['Authz_Conf_File'], $this->local_site);
    		if (!$authz->authorize($user, $this->create_Hash(1,$this->assertion), $_GET['ref'], $force)) {
    			$this->log("error al autorizar: ".$authz->getError());
    			return array(false, $this->cfg['Invite_Error_File']);
    		}
    	}
    	$this->log("authorizeWrapper(): ".$user." is now authorized.");
    	return array(true, null);
	}

	function invite($mail,$lang, $url) {
		if (isset($this->cfg['Authz_Engine'])) {
			include_once($this->cfg['Authz_Engine'].".php");
			$authz = new $this->cfg['Authz_Engine']();
    		$authz->configure($this->cfg['Authz_Conf_File'], $this->local_site);
			if (!$authz->invite($mail,$lang,$url)) {
				$this->log($authz->getError());
				return false;
			}
		}
		$this->log("invite(): invite sent to ".$mail.".");
		return true;
	}

    function removeInvite($ref) {
        if (isset($this->cfg['Authz_Engine'])) {
            include_once($this->cfg['Authz_Engine'].".php");
            $authz = new $this->cfg['Authz_Engine']();
            $authz->configure($this->cfg['Authz_Conf_File'], $this->local_site);
            if (!$authz->removeInvite($ref)) {
                $this->log($authz->getError());
                return false;
            }
        }
        $this->log("removeInvite(): invite ".$ref." removed.");
        return true;
    }

	function revoke($mail) {
		if (isset($this->cfg['Authz_Engine'])) {
			include_once($this->cfg['Authz_Engine'].".php");
			$authz = new $this->cfg['Authz_Engine']();
			$authz->configure($this->cfg['Authz_Conf_File'], $this->local_site);
			if (!$authz->revoke($mail)) {
				$this->log($authz->getError());
				return false;
			}
		}
		$this->log("revoke(): ".$mail." authorization is now revoked.");
		return true;
	}

	function getAuthorizedList() {
		if (isset($this->cfg['Authz_Engine'])) {
			include_once($this->cfg['Authz_Engine'].".php");
			$authz = new $this->cfg['Authz_Engine']();
			$authz->configure($this->cfg['Authz_Conf_File'], $this->local_site);
			return $authz->getAuthorizedList();
		}
		return false;
	}

	function getPendingInvites() {
		if (isset($this->cfg['Authz_Engine'])) {
			include_once($this->cfg['Authz_Engine'].".php");
			$authz = new $this->cfg['Authz_Engine']();
			$authz->configure($this->cfg['Authz_Conf_File'], $this->local_site);
			return $authz->getPendingInvites();
		}
		return false;
	}

    function logout() {
        setcookie($this->Lcook_name,$_COOKIE[$this->Lcook_name],time()-3600);
    }
} //class abstract PoA

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS simplePoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////
abstract class simplePoA extends abstractPoA {
	var $config_file;

	function set_parameters() {
		// Get config file name from php.ini
		$phpPoA_ini_file = get_cfg_var('phpPoA_ini_file');
		return $this->set_Global_Parameters($phpPoA_ini_file);
	}
	

/* REDUNDANT!
    function test_Filters($userAssertion) {
		$accept = "";
		$reject = "";

		// Main filter
		if(isset($this->config_file["PAPI_Main"]["PAPI_Filter_accept"])){
			$accept = $this->config_file["PAPI_Main"]["PAPI_Filter_accept"];
		}
		if(isset($this->config_file["PAPI_Main"]["PAPI_Filter_reject"])){
			$reject = $this->config_file["PAPI_Main"]["PAPI_Filter_reject"];
		}
		$main_match = $this->match_Filters($userAssertion, $accept, $reject);

		// aggregated filters
		if(isset($this->cfg["PAPI_Filter_accept"])){
			$accept = $this->cfg["PAPI_Filter_accept"];
		}
		if(isset($this->cfg["PAPI_Filter_reject"])){
			$reject = $this->cfg["PAPI_Filter_reject"];
		}
		$local_match = $this->match_Filters($userAssertion, $accept, $reject);

		// Global evaluation
		if ($local_match) {// && $main_match) {
			return 1;
		} else {
			$error_message = "test_Filters() Function: Code $userAssertion matches assertion rejection filter.";
			$this->log($error_message);
			return 0;
		}
	}
*/
	function send_Request() {
		$request = $this->prepare_Request();
		if (!$request) {
			# If AutoPoA automatically redirect error page
			# If PoA return userAssertion
			return array(-1, "", "Location: " . $this->cfg['System_Error_File']);
		}
		header("HTTP/1.1 302 Found");
		header("Location: $request");
		exit;
	}

	function set_Global_Parameters($phpPoA_ini_file) {
		// Get parameters from config file
		$this->config_file = parse_ini_file($phpPoA_ini_file, true);
		// PAPI_Main Section
		$MAIN_Params = Array (
			"Lcook_Timeout",
			"Request_DB",
			"DB_Type",
			"Log",
			"Not_Auth_Error_File",
			"Cookie_Error_File",
			"System_Error_File",
			"Config_Error_File",
			"Invite_Error_File",
			"PAPI_Filter_accept",
			"PAPI_Filter_reject",
			"Pass_Pattern",
			"Cookie_Domain",
			"LKEY_File",
			"GPoA_Pub_Key",
			"GPoA_URL",
			"AS_Pub_Key",
			"AS_URL",
			"Allow_From",
			"Deny_From",
			"Authz_Engine",
			"Authz_Conf_File",
			"MYSQL_SERVER",
			"MYSQL_USER",
			"MYSQL_PASS",
			"MYSQL_DB",
			"Home_Locator"
		);
		if (!isset ($this->config_file["PAPI_Main"]["Cookie_Domain"])) {
			$this->config_file["PAPI_Main"]["Cookie_Domain"] = $_SERVER["SERVER_NAME"];
		}
		foreach ($MAIN_Params as $param) {
			if(isset($this->config_file["PAPI_Main"][$param])){
				$this->cfg[$param] = $this->config_file["PAPI_Main"][$param];
			}
		}
		// The parameters in local section of phpPoA_ini_file have priority over parameters in the
		// PAPI_Main section with the same name. Therefore the local parameters set their own value
		// over global parameters. As the filter parameters must not have priority because we have
		// to check both of them, to access PAPI_Filter_accept and PAPI_Filter_reject from PAPI_Main
		// Section we have to use:
		// $config_file["PAPI_Main"]["PAPI_Filter_accept"]
		// $config_file["PAPI_Main"]["PAPI_Filter_reject"]
		// instead of $cfg.
		// PAPI_Local Section
		$LOCAL_Params = Array (
			"Location",
			"Lcook_Timeout",
			"Request_DB",
			"DB_Type",
			"Log",
			"Not_Auth_Error_File",
			"Cookie_Error_File",
			"System_Error_File",
			"Config_Error_File",
			"Invite_Error_File",
			"Cookie_Domain",
			"PAPI_Filter_accept",
			"PAPI_Filter_reject",
			"Pass_Pattern",
			"LKEY_File",
			"GPoA_Pub_Key",
			"GPoA_URL",
			"AS_Pub_Key",
			"AS_URL",
			"Allow_From",
			"Deny_From",
			"Authz_Engine",
			"Authz_Conf_File",
			"MYSQL_SERVER",
			"MYSQL_USER",
			"MYSQL_PASS",
			"MYSQL_DB",
			"Home_Locator"
		);
		if (!isset ($this->config_file[$this->local_site]["Cookie_Domain"])) {
			$this->config_file[$this->local_site]["Cookie_Domain"] = $_SERVER["SERVER_NAME"];
		}
		foreach ($LOCAL_Params as $param) {
			if (isset($this->config_file[$this->local_site][$param])) {
			    if (is_array($this->config_file[$this->local_site][$param])) {
			        if (!is_array($this->cfg[$param])) {
			            $this->cfg[$param] = array($this->cfg[$param]);
			        }
			        $this->cfg[$param] = array_merge($this->cfg[$param],
			                                             $this->config_file[$this->local_site][$param]);
			    } else {
				    $this->cfg[$param] = $this->config_file[$this->local_site][$param];
			    }
			}
		}
		return True;
	} // set_Global_Parameters()

	function log($message) {
		$error_message = "[".date("D M d H:i:s Y")."] [".$this->local_site."] [client ".$_SERVER['REMOTE_ADDR']."] ".$message;
		error_log($error_message."\n",3,$this->cfg["Log"]);
	}
} //class simplePoA

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS PoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class PoA extends simplePoA {

	function check_Access() {
		list($code, $this->assertion) = $this->papi_Engine();
		$userData = $this->create_Hash($code, $this->assertion);
		return $userData;
	}

	function isAuthorized($user) {
		list($code, $location) = $this->isAuthorizedWrapper($user);
		return $code;
	}

	function authorize($user, $force = false) {
		list($code, $location) = $this->authorizeWrapper($user, $force);
		return $code;
	}
} // Class PoA


////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS autoPoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class autoPoA extends simplePoA {
	function check_Access() {
		list($code, $this->assertion, $location) = $this->papi_Engine();
		if ($location && $code < 1) // error found, redirect!
			header("Location: " . $location);
		else // this shouldn't be an error, but can be. Anyway, do not redirect!
			return $this->create_Hash($code, $this->assertion);
	}

	function isAuthorized($user) {
		list($code, $location) = $this->isAuthorizedWrapper($user);
		if (!$code) { // error found, redirect!
			header("Location: ".$location);
		}
		return $code;
	}

	function authorize($user, $force = false) {
		list($code, $location) = $this->authorizeWrapper($user, $force);
		if (!$code) { // error found, redirect!
			header("Location: ".$location);
		}
		return $code;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CLASS thinPoA
////////////////////////////////////////////////////////////////////////////////////////////////////////////////

class thinPoA extends autoPoA {
    var $send_request = true;
}
?>
